import struct
import sys

import mock
import pytest


class MockSerial:
    def __init__(self, *args, **kwargs):
        self.ptr = 0
        self.sof = b"\x42\x4d"
        self.data = self.sof
        self.data += struct.pack(">H", 28)
        self.data += b"\x00" * 26
        checksum = struct.pack(">H", sum(bytearray(self.data)))
        self.data += checksum

    def read(self, length):
        result = self.data[self.ptr : self.ptr + length]
        self.ptr += length
        if self.ptr >= len(self.data):
            self.ptr = 0
        return result

    def flushInput(self):
        pass

    def close(self):
        pass


class MockSerialFail(MockSerial):
    def __init__(self, *args, **kwargs):
        pass

    def read(self, length):
        return b"\x00" * length


@pytest.fixture(scope='function', autouse=False)
def pms5003():
    import pms5003
    yield pms5003
    del sys.modules['pms5003']


@pytest.fixture(scope='function', autouse=False)
def gpiod():
    sys.modules['gpiod'] = mock.Mock()
    sys.modules['gpiod.line'] = mock.Mock()
    yield sys.modules['gpiod']
    del sys.modules['gpiod.line']
    del sys.modules['gpiod']


@pytest.fixture(scope='function', autouse=False)
def gpiodevice():
    gpiodevice = mock.Mock()
    gpiodevice.get_pins_for_platform.return_value = [(mock.Mock(), 0), (mock.Mock(), 0)]
    gpiodevice.get_pin.return_value = (mock.Mock(), 0)

    sys.modules['gpiodevice'] = gpiodevice
    yield gpiodevice
    del sys.modules['gpiodevice']


@pytest.fixture(scope='function', autouse=False)
def serial():
    sys.modules['serial'] = mock.Mock()
    sys.modules['serial'].Serial = MockSerial
    yield sys.modules['serial']
    del sys.modules['serial']


@pytest.fixture(scope='function', autouse=False)
def serial_fail():
    sys.modules['serial'] = mock.Mock()
    sys.modules['serial'].Serial = MockSerialFail
    yield sys.modules['serial']
    del sys.modules['serial']




import struct
import time

import gpiod
import gpiodevice
import serial
from gpiod.line import Direction, Value

__version__ = "1.0.1"


PMS5003_SOF = bytearray(b"\x42\x4d")

OUTL = gpiod.LineSettings(direction=Direction.OUTPUT, output_value=Value.INACTIVE)
OUTH = gpiod.LineSettings(direction=Direction.OUTPUT, output_value=Value.ACTIVE)


class ChecksumMismatchError(RuntimeError):
    pass


class ReadTimeoutError(RuntimeError):
    pass


class SerialTimeoutError(RuntimeError):
    pass


class PMS5003Data:
    def __init__(self, raw_data):
        self.raw_data = raw_data
        self.data = struct.unpack(">HHHHHHHHHHHHHH", raw_data)
        self.checksum = self.data[13]

    def pm_ug_per_m3(self, size, atmospheric_environment=False):
        if atmospheric_environment:
            if size == 1.0:
                return self.data[3]
            if size == 2.5:
                return self.data[4]
            if size is None:
                return self.data[5]

        else:
            if size == 1.0:
                return self.data[0]
            if size == 2.5:
                return self.data[1]
            if size == 10:
                return self.data[2]

        raise ValueError("Particle size {} measurement not available.".format(size))

    def pm_per_1l_air(self, size):
        if size == 0.3:
            return self.data[6]
        if size == 0.5:
            return self.data[7]
        if size == 1.0:
            return self.data[8]
        if size == 2.5:
            return self.data[9]
        if size == 5:
            return self.data[10]
        if size == 10:
            return self.data[11]

        raise ValueError("Particle size {} measurement not available.".format(size))

    def __repr__(self):
        return """
PM1.0 ug/m3 (ultrafine particles):                             {}
PM2.5 ug/m3 (combustion particles, organic compounds, metals): {}
PM10 ug/m3  (dust, pollen, mould spores):                      {}
PM1.0 ug/m3 (atmos env):                                       {}
PM2.5 ug/m3 (atmos env):                                       {}
PM10 ug/m3 (atmos env):                                        {}
>0.3um in 0.1L air:                                            {}
>0.5um in 0.1L air:                                            {}
>1.0um in 0.1L air:                                            {}
>2.5um in 0.1L air:                                            {}
>5.0um in 0.1L air:                                            {}
>10um in 0.1L air:                                             {}
""".format(
            *self.data[:-2]
        )

    def __str__(self):
        return self.__repr__()


class PMS5003:
    def __init__(self, device="/dev/ttyAMA0", baudrate=9600, pin_enable="GPIO22", pin_reset="GPIO27"):
        self._serial = None
        self._device = device
        self._baudrate = baudrate

        gpiodevice.friendly_errors = True
        self._pin_enable = gpiodevice.get_pin(pin_enable, "PMS5003_en", OUTH)
        self._pin_reset = gpiodevice.get_pin(pin_reset, "PMS5003_rst", OUTL)

        self.setup()

    def setup(self):
        if self._serial is not None:
            self._serial.close()

        self._serial = serial.Serial(self._device, baudrate=self._baudrate, timeout=4)

        self.reset()

    def set_pin(self, pin, state):
        lines, offset = pin
        lines.set_value(offset, Value.ACTIVE if state else Value.INACTIVE)

    def reset(self):
        time.sleep(0.1)
        self.set_pin(self._pin_reset, False)
        self._serial.flushInput()
        time.sleep(0.1)
        self.set_pin(self._pin_reset, True)

    def read(self):
        start = time.time()

        sof_index = 0

        while True:
            elapsed = time.time() - start
            if elapsed > 5:
                raise ReadTimeoutError("PMS5003 Read Timeout: Could not find start of frame")

            sof = self._serial.read(1)
            if len(sof) == 0:
                raise SerialTimeoutError("PMS5003 Read Timeout: Failed to read start of frame byte")
            sof = ord(sof) if isinstance(sof, bytes) else sof

            if sof == PMS5003_SOF[sof_index]:
                if sof_index == 0:
                    sof_index = 1
                elif sof_index == 1:
                    break
            else:
                sof_index = 0

        checksum = sum(PMS5003_SOF)

        data = bytearray(self._serial.read(2))  # Get frame length packet
        if len(data) != 2:
            raise SerialTimeoutError("PMS5003 Read Timeout: Could not find length packet")
        checksum += sum(data)
        frame_length = struct.unpack(">H", data)[0]

        raw_data = bytearray(self._serial.read(frame_length))
        if len(raw_data) != frame_length:
            raise SerialTimeoutError("PMS5003 Read Timeout: Invalid frame length. Got {} bytes, expected {}.".format(len(raw_data), frame_length))

        data = PMS5003Data(raw_data)
        # Don't include the checksum bytes in the checksum calculation
        checksum += sum(raw_data[:-2])

        if checksum != data.checksum:
            raise ChecksumMismatchError("PMS5003 Checksum Mismatch {} != {}".format(checksum, data.checksum))

        return data



import pytest


def test_setup(gpiod, gpiodevice, serial, pms5003):
    _ = pms5003.PMS5003()


def test_double_setup(gpiod, gpiodevice, serial, pms5003):
    sensor = pms5003.PMS5003()
    sensor.setup()


def test_read(gpiod, gpiodevice, serial, pms5003):
    sensor = pms5003.PMS5003()
    data = sensor.read()
    data.pm_ug_per_m3(2.5)


def test_read_fail(gpiod, gpiodevice, serial_fail, pms5003):
    sensor = pms5003.PMS5003()
    with pytest.raises(pms5003.ReadTimeoutError):
        _ = sensor.read()



#!/usr/bin/env python

from pms5003 import PMS5003

print(
    """specific.py - Continuously print a specific data value.

Press Ctrl+C to exit!

"""
)

# Configure the PMS5003 for Enviro+
# pins and ports may vary for your hardware!

# Default, assume Raspberry Pi compatible, running Raspberry Pi OS Bookworm
pms5003 = PMS5003(device="/dev/ttyAMA0", baudrate=9600)

# Raspberry Pi 4 (Raspberry Pi OS)
#
# GPIO22 and GPIO27 are enable and reset for Raspberry Pi 4
# use "raspi-config" to enable serial, or add
# "dtoverlay=uart0" to /boot/config.txt
#
# pms5003 = PMS5003(device="/dev/ttyAMA0", baudrate=9600, pin_enable="GPIO22", pin_reset="GPIO27")

# Raspberry Pi 5 (Raspberry Pi OS)
#
# GPIO22 and GPIO27 are enable and reset for Raspberry Pi 5
# On older versions of Bookworm these might be PIN15 and PIN13
# use "raspi-config" to enable serial, or add
# "dtoverlay=uart0-pi5" to /boot/firmware/config.txt
#
# pms5003 = PMS5003(device="/dev/ttyAMA0", baudrate=9600, pin_enable="GPIO22", pin_reset="GPIO27")

# ROCK 5B
#
# Use "armbian-config" to enable rk3568-uart2-m0
# Disable console on ttyS2 with:
# sudo systemctl stop serial-getty@ttyS2.service
# sudo systemctl disable serial-getty@ttyS2.service
# sudo systemctl mask serial-getty@ttyS2.service
# add "console=display" to /boot/armbianEnv.txt
#
# pms5003 = PMS5003(device="/dev/ttyS2", baudrate=9600, pin_enable="PIN_15", pin_reset="PIN_13")

# Other
#
# Use gpiod to request the pins you want, and pass those into PMS5003 as LineRequest, offset tuples.
#
# from pms5003 import OUTL, OUTH
# from gpiod import Chip
# lines = Chip.request_lines(consumer="PMS5003", config={22: OUTH, 27: OUTL})
# pms5003 = PMS5003(device="/dev/ttyAMA0", baudrate=9600, pin_enable=(lines, 22), pin_reset=(lines, 27))


try:
    while True:
        data = pms5003.read()
        print(f"PM2.5 ug/m3 (combustion particles, organic compounds, metals): {data.pm_ug_per_m3(2.5)}")

except KeyboardInterrupt:
    pass
